/* 
richards.c
calculation of soil water content layer by layer taking into account soil hydrological processes 
(percolation, diffusion)  using inamic discretization level in order to taking account the change of the diffusivity and conductivity depending on soil moisture 
(after rain event the soil moisture rises suddenly causing high diffusivity and conductivity. If we assumed that these diffusivity 
 and conductivity values were valid all the day we would overestimate the velocity diffusion and percolation process) 

*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
Biome-BGCMuSo v6.2.
Copyright 2020, D. Hidy [dori.hidy@gmail.com]
Hungarian Academy of Sciences, Hungary
See the website of Biome-BGCMuSo at http://nimbus.elte.hu/bbgc/ for documentation, model executable and example input files.
*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <malloc.h>
#include "ini.h"
#include "bgc_struct.h"
#include "bgc_constants.h"
#include "bgc_func.h"
#define MIN(X, Y) (((X) < (Y)) ? (X) : (Y))
#define MAX(X, Y) (((X) > (Y)) ? (X) : (Y))   

int richards(siteconst_struct* sitec, soilprop_struct* sprop, const epconst_struct* epc, epvar_struct* epv, wstate_struct* ws, wflux_struct* wf)
{

	
	/* internal variables */
	int errorCode;
	double INFILT,INFILT_act,INFILT_ctrl,EVAP,EVAP_act,EVAP_ctrl,transpDEM,transpDEM_act;
	double VWCequilib,diffus_limit;

	/* diffusion and percolation calculation */
	double D0, D1, Dact;	        /* hydrological diffusion coefficient (m2/s) */

	int layer, n_second, n_sec, discretlevel;

	double VWCdiff_max, VWCdiff;
	double localvalue,  exponent_discretlevel;
	double soilw0, soilw1, VWC0, VWC1, soilw_sat0, soilw_fc0, soilw_wp0, soilw_hw0;
	double VWCsat0, VWCsat1, Ksat0, Ksat1, Dsat0, Dsat1, soilB0, soilB1, PSIsat0, PSIsat1, Kact;

	double evap_lack,transp_lack, pondw_act;
	double soilw_to_pondw, infilt_to_soilw, pondw_evap, soilw_evap, transpDEM_layer;

	double dz0, dz1;
	double transp[N_SOILLAYERS],diffus[N_SOILLAYERS], percol[N_SOILLAYERS],transp_ctrl[N_SOILLAYERS],diffus_ctrl[N_SOILLAYERS], percol_ctrl[N_SOILLAYERS];
	errorCode = 0;

	INFILT=EVAP=transpDEM=INFILT_ctrl=EVAP_act=EVAP_ctrl=transpDEM_layer=transpDEM_act=0;
	evap_lack=transp_lack=soilw_to_pondw=infilt_to_soilw=pondw_evap=soilw_evap=pondw_act=0;
	for (layer=0 ; layer < N_SOILLAYERS; layer++)
	{
		transp_ctrl[layer]=0;
		diffus_ctrl[layer]=0;
		percol_ctrl[layer]=0;
		transp[layer]=0;
		diffus[layer]=0;
		percol[layer]=0;
	}
	ws->iterstep_Richards=0;

	/* parameter values balus
		int inORout, fineRESact;
		
	richards_struct rich;
	rich.critDepth = 1;
	rich.dz_aboveCD = 0.01;
	rich.dz_belowCD = 0.1;

	if (epc->SHCM_flag== 3)
		fineRESact = fineRES;
	else
		fineRESact = N_SOILLAYERS;  */
	/* --------------------------------------------------------------------------------------------------------*/
	/* I. INITALIZATION */
	
	n_second=0;
	n_sec=1;
	discretlevel = epc->discretlevel_Richards + 3; /* discretization level: change in VWC 1% -> timestep: second */

	
	INFILT     = (wf->prcp_to_soilw + wf->snoww_to_soilw + wf->canopyw_to_soilw + wf->IRG_to_prcp);
	INFILT_act = INFILT/nSEC_IN_DAY;

	EVAP         = wf->soilw_evapPOT;
	EVAP_act     = EVAP/nSEC_IN_DAY;
	transpDEM     = wf->soilw_transDEMAND_SUM;
	transpDEM_act = transpDEM/nSEC_IN_DAY;

	pondw_act = ws->pondw;
	
	/* --------------------------------------------------------------------------------------------------------*/	
	/* II. CALCULATE PROCESSES WITH DYNAMIC TIME STEP LAYER TO LAYER (except of bottom layer) */

	while (n_second < nSEC_IN_DAY)
	{
	
		VWCdiff_max=0;

		/* 1. CALCULATE PROCESSES  LAYER TO LAYER */
		for (layer=0 ; layer < N_SOILLAYERS-1; layer++)
		{	
			/* 1.0. INITALIZATION */
			dz0        = sitec->soillayer_thickness[layer];
			dz1        = sitec->soillayer_thickness[layer+1];
			soilw0     = ws->soilw[layer];
			soilw1     = ws->soilw[layer+1];
			VWC0       = epv->VWC[layer];
			VWC1       = epv->VWC[layer+1];
			VWCsat0    = sprop->VWCsat[layer];
			VWCsat1    = sprop->VWCsat[layer+1];
			soilw_sat0 = sprop->VWCsat[layer] * dz0 * water_density;
			soilw_fc0  = sprop->VWCfc[layer]* dz0 * water_density;
			soilw_wp0  = sprop->VWCwp[layer] * dz0 * water_density;
			soilw_hw0  = sprop->VWChw[layer] * dz0 * water_density;
		
			Ksat0 = sprop->hydrCONDUCTsat[layer];
			Ksat1 = sprop->hydrCONDUCTsat[layer+1];
			Dsat0 = sprop->hydrDIFFUSsat[layer];
			Dsat1 = sprop->hydrDIFFUSsat[layer+1];

			soilB0 = sprop->soilB[layer];
			soilB1 = sprop->soilB[layer+1];

			PSIsat0 = sprop->PSIsat[layer];
			PSIsat1 = sprop->PSIsat[layer+1];
			
			/* -----------------------------*/
			/* 1.1. INFILTRATION */
  			if (layer == 0) 
			{

				if ((soilw0 + INFILT_act - soilw_sat0) > CRIT_PREC)			
				{
					soilw_to_pondw      = (soilw0 + INFILT_act - soilw_sat0); //balus
					INFILT_act         -= (soilw0 + INFILT_act - soilw_sat0);
					INFILT             -= (soilw0 + INFILT_act - soilw_sat0);
					infilt_to_soilw     = soilw_sat0 - soilw0;
				}
				else
					infilt_to_soilw     = INFILT_act;

				INFILT           -= INFILT_act;
				INFILT_ctrl       += INFILT_act;
			}

			/* -----------------------------*/
			/* 1.2. EVAPORATION */
  			if (layer == 0) 
			{
				/* potential evaporation < pondw -> only pondw is evaporated */
				if (EVAP_act < pondw_act)
				{
					pondw_evap = EVAP_act;
					soilw_evap = 0;
				}
				else
				{
					/* potential evaporation > pondw but pondw exists -> both pondw and soilw are evaporated */
					pondw_evap = pondw_act;
						
					/* available SWC can limit evaporation: evap_lack = evap - soilw_avail -> if evap_lack is positive: evaporation limitation */
					evap_lack = (EVAP_act - pondw_evap) - (soilw0 - soilw_hw0);
					
					if (evap_lack > 0) EVAP_act -= evap_lack;
		
					soilw_evap = EVAP_act - pondw_evap;
				}
				
				EVAP           -= EVAP_act;
				EVAP_ctrl      += EVAP_act;

			}

			/* -----------------------------*/
			/* 1.3. transpIRATION */

			/* transpiration based on rootlenght proportion */
			transpDEM_layer = transpDEM_act * epv->rootlength_prop[layer]; 

			/* transp_lack: control parameter to avoid negative soil water content (due to overestimated transpiration + dry soil) */
			transp_lack = transpDEM_layer - (soilw0 - soilw_wp0);

			/* if transpiration demand is greater than theoretical lower limit of water content: wilting point -> limited transpiration flux)  */
			if (transp_lack > 0) 
				transp[layer] = transpDEM_layer - transp_lack;
			else
				transp[layer] = transpDEM_layer;

			
			transp_ctrl[layer] += transp[layer];
	
			transpDEM           -= transp[layer];

			/* -----------------------------*/
			/* 1.4. PERCOLATION */

			/* conductivity coefficient - theoretical upper limit: saturation value */
			Kact = Ksat0 * pow(VWC0/sprop->VWCsat[layer], 2*(sprop->soilB[layer])+3);


			/* percolation flux */
			percol[layer] = Kact * water_density * n_sec; // kg/m2
			percol_ctrl[layer] = percol[layer];
			

			/* -----------------------------*/
			/* 1.5. DIFFUSION */
		
			/* diffusivity coefficient 	*/
			D0 = (((soilB0  * Ksat0   * (-100*PSIsat0)))   / VWC0) * pow(VWC0/VWCsat0,  soilB0+2);
			
			D1 = (((soilB1  * Ksat1   * (-100*PSIsat1)))   / VWC1) * pow(VWC1/VWCsat1,  soilB1+2);

			Dact = (D0 * dz0/(dz0+dz1) + D1 * dz1/(dz0+dz1));
	
		
			/* diffusion flux  */
			diffus[layer]  = (VWC0 - VWC1)/((dz0+dz1)/2) 	* Dact  * water_density * n_sec;


			VWCequilib = (VWC0 * dz0 + VWC1 * dz1) / (dz0+dz1);
			
			diffus_limit = (VWCequilib - VWC1) * dz1 * water_density;

			if (fabs(diffus[layer]) > CRIT_PREC && fabs(diffus[layer]) > fabs(diffus_limit))
			{
				diffus[layer]  = diffus_limit;
			}

			diffus_ctrl[layer] += diffus[layer];
			
		}

		/* 2. UPDATING */
		for (layer=0 ; layer < N_SOILLAYERS; layer++)
		{
			/* 2.1 STATE VARIABLES */
			if (layer == 0)
				ws->soilw[layer] = ws->soilw[layer] + infilt_to_soilw - soilw_to_pondw - soilw_evap - transp[layer] - percol[layer] - diffus[layer];
			else
				ws->soilw[layer] = ws->soilw[layer] - transp[layer] + percol[layer-1] + diffus[layer-1] - percol[layer] - diffus[layer];
			
			VWC0 = epv->VWC[layer];
			epv->VWC[layer] = ws->soilw[layer] / (water_density * sitec->soillayer_thickness[layer]);
			
			/* 2.2 FLUX VARIABLES */
			if (layer == 0)
			{
				wf->soilw_to_pondw += soilw_to_pondw;
				wf->soilw_evap     += soilw_evap;
			}

			wf->soilw_transp[layer]     += transp[layer];
			wf->soilw_percolated[layer] += percol[layer];
			wf->soilw_diffused[layer]   += diffus[layer];
		
			/* -----------------------------*/
			/* 2.3. FINDING THE MAXIMAL CHANGE to determine discretization step */
			
			VWCdiff = fabs(VWC0 - epv->VWC[layer]);
			
			if (VWCdiff_max < VWCdiff) VWCdiff_max = VWCdiff;		
		
		
		}
			
		n_second += n_sec;

		/* ********************************/	
		/* 4.2. CALCULATION OF DISCRETE LEVEL: if the maximal change of VWC is not equal to 0 (greater than) -> discretlevel is the function of local value of the change */
		VWCdiff_max = VWCdiff_max/n_sec;
		if (VWCdiff_max > CRIT_PREC)
		{
			localvalue=floor(log10(VWCdiff_max));
			if (localvalue + discretlevel < 0)
				exponent_discretlevel = fabs(localvalue + discretlevel);
			else
				exponent_discretlevel = 0;

			n_sec = (int) pow(10, exponent_discretlevel);
			if (n_sec + n_second > nSEC_IN_DAY) 
				n_sec = nSEC_IN_DAY - n_second;
		}
		else
		{
			n_sec = nSEC_IN_DAY - n_second;
		}
		VWCdiff_max = 0;
		if (n_second < nSEC_IN_DAY)
		{
			INFILT_act = INFILT/(nSEC_IN_DAY-n_second) * n_sec;
			EVAP_act = EVAP/(nSEC_IN_DAY-n_second) * n_sec;
			transpDEM_act = transpDEM/(nSEC_IN_DAY-n_second) * n_sec;
		}
		else
		{
			INFILT_act = 0;
			EVAP_act = 0;
			transpDEM_act = 0;
		}


		ws->iterstep_Richards=ws->iterstep_Richards+1;	
	}

	/* change 140 layers to 10 layers 
	inORout=1;
	if (richards_fineRES(fineRESact, inORout, epc,sitec, sprop, epv, ws, wf, &rich))
	{
		if (!errorCode) 
		{
			printf("\n");
			printf("ERROR: richards_fineRES() for richards \n");
		}
		errorCode=1;
	}	
	*/
	return (errorCode);
}


int richards_fineRES(int fineRESact, int inORout, const epconst_struct* epc, siteconst_struct* sitec, soilprop_struct* sprop, epvar_struct* epv, 
	                 wstate_struct* ws, wflux_struct* wf, richards_struct* rich)
{	
	int nlayer, fr, errorCode, layerbound;
	double depth, dz;
	nlayer=errorCode=layerbound = 0;
	
	if (epc->SHCM_flag == 3)
	{
		/* new VWC values after Richards-method */
		if (inORout == 1)
		{
			for (nlayer = 0; nlayer < N_SOILLAYERS-1; nlayer++) ws->soilw[nlayer] = 0;
			nlayer=0;
		}

		/*  depth of the first layer in fine resolution */
		depth=rich->dz_aboveCD;  


		for (fr=0 ; fr < fineRESact-1; fr++)
		{
			if (depth - rich->critDepth > CRIT_PREC)
				dz=rich->dz_belowCD;
			else
				dz=rich->dz_aboveCD;

			if (fabs(depth - sitec->soillayer_depth[nlayer]) < CRIT_PREC)
				layerbound=1;
			else
				layerbound=0;
		
	
			/* input for Richards */
			if (inORout == 0)
			{
				rich->VWC_FR[fr]      = epv->VWC[nlayer];
				rich->VWCsat_FR[fr]   = sprop->VWCsat[nlayer];
				rich->VWCfc_FR[fr]    = sprop->VWCfc[nlayer];
				rich->VWCwp_FR[fr]    = sprop->VWCwp[nlayer];
				rich->VWChw_FR[fr]    = sprop->VWChw[nlayer];
				rich->PSIsat_FR[fr]   = sprop->PSIsat[nlayer];
				rich->soilB_FR[fr]    = sprop->soilB[nlayer];
				rich->Ksat_FR[fr]     = sprop->hydrCONDUCTsat[nlayer];
				rich->Dsat_FR[fr]     = sprop->hydrDIFFUSsat[nlayer];
				rich->Kact_FR[fr]     = epv->hydrCONDUCTact[nlayer];
				rich->Dact_FR[fr]     = epv->hydrDIFFUSact[nlayer];
				rich->percol_FR[fr]   = 0;
				rich->diffus_FR[fr]   = 0;
				rich->soilw_FR[fr]    = rich->VWC_FR[fr]*dz*water_density;
				rich->dz_FR[fr]       = dz;
			}

			/* output from Richards */
			else
			{
				/* averaged VWC */
				ws->soilw[nlayer]     += rich->soilw_FR[fr];

				/* percolation and diffusion from boundary fineRESact layer */
				if (layerbound)
				{
					wf->soilw_percolated[nlayer] = rich->percol_FR[fr];
					wf->soilw_diffused[nlayer] = rich->diffus_FR[fr];
					epv->hydrCONDUCTact[nlayer] = rich->Kact_FR[fr];
					epv->hydrDIFFUSact[nlayer] = rich->Dact_FR[fr];
				}
			}

	
			if (layerbound) nlayer=nlayer+1;	


			if (fabs(depth - rich->critDepth) < CRIT_PREC)
				depth+=rich->dz_belowCD;
			else
				depth+=dz;
		
		}
	}
	else
	{
		for (fr=0 ; fr < fineRESact-1; fr++)
		{
			/* input for Richards */
			if (inORout == 0)
			{
				rich->VWC_FR[fr]      = epv->VWC[fr];
				rich->VWCsat_FR[fr]   = sprop->VWCsat[fr];
				rich->VWCfc_FR[fr]    = sprop->VWCfc[fr];
				rich->VWCwp_FR[fr]    = sprop->VWCwp[fr];
				rich->VWChw_FR[fr]    = sprop->VWChw[fr];
				rich->PSIsat_FR[fr]   = sprop->PSIsat[fr];
				rich->soilB_FR[fr]    = sprop->soilB[fr];
				rich->Ksat_FR[fr]     = sprop->hydrCONDUCTsat[fr];
				rich->Dsat_FR[fr]     = sprop->hydrDIFFUSsat[fr];
				rich->Kact_FR[fr]     = epv->hydrCONDUCTact[fr];
				rich->Dact_FR[fr]     = epv->hydrDIFFUSact[fr];
				rich->percol_FR[fr]   = 0;
				rich->diffus_FR[fr]   = 0;
				rich->soilw_FR[fr]    = ws->soilw[fr];
				rich->dz_FR[fr]       = sitec->soillayer_thickness[fr];
			}
			else
			{
			
								/* averaged VWC */
				ws->soilw[fr]     = rich->soilw_FR[fr];
				epv->VWC[fr]     = rich->VWC_FR[fr];
				wf->soilw_percolated[fr] = rich->percol_FR[fr];
				wf->soilw_diffused[fr] = rich->diffus_FR[fr];
				epv->hydrCONDUCTact[fr] = rich->Kact_FR[fr];
				epv->hydrDIFFUSact[fr] = rich->Dact_FR[fr];

			}
		}
	
	}

	
	/* input for Richards:  bottom layer is special -> no VWC change */
	if (inORout == 0)
	{
		dz                              = sitec->soillayer_thickness[N_SOILLAYERS-1];
		rich->VWC_FR[fineRESact-1]      = epv->VWC[N_SOILLAYERS-1];
		rich->VWCfc_FR[fineRESact-1]    = sprop->VWCfc[N_SOILLAYERS-1];
		rich->VWCwp_FR[fineRESact-1]    = sprop->VWCwp[N_SOILLAYERS-1];
		rich->VWChw_FR[fineRESact-1]    = sprop->VWChw[N_SOILLAYERS-1];
		rich->VWCsat_FR[fineRESact-1]   = sprop->VWCsat[N_SOILLAYERS-1];
		rich->PSIsat_FR[fineRESact-1]   = sprop->PSIsat[N_SOILLAYERS-1];
		rich->soilB_FR[fineRESact-1]    = sprop->soilB[N_SOILLAYERS-1];
		rich->Ksat_FR[fineRESact-1]     = sprop->hydrCONDUCTsat[N_SOILLAYERS-1];
		rich->Dsat_FR[fineRESact-1]     = sprop->hydrDIFFUSsat[N_SOILLAYERS-1];
		rich->Kact_FR[fineRESact-1]     = epv->hydrCONDUCTact[N_SOILLAYERS-1];
		rich->Dact_FR[fineRESact-1]     = epv->hydrDIFFUSact[N_SOILLAYERS-1];
		rich->percol_FR[fineRESact-1]   = 0;
		rich->diffus_FR[fineRESact-1]   = 0;
		rich->soilw_FR[fineRESact-1]    = rich->VWC_FR[N_SOILLAYERS-1]*dz*water_density;
		rich->dz_FR[fineRESact-1]       = dz;
	}

	/* calculation of VWC from soilw */
	if (inORout == 1)
	{
		for (nlayer = 0; nlayer < N_SOILLAYERS-1; nlayer++)  epv->VWC[nlayer] = ws->soilw[nlayer] / (water_density * sitec->soillayer_thickness[nlayer]);
		/* bottom layer is special -> no VWC change */
		wf->soilw_percolated[N_SOILLAYERS-1] = wf->soilw_percolated[N_SOILLAYERS-2];
		wf->soilw_diffused[N_SOILLAYERS-1]   = wf->soilw_diffused[N_SOILLAYERS-2];
	}



	


	return (errorCode);
}